#!/usr/bin/python
# -*- coding: utf-8 -*-

# Programming contest management system
# Copyright © 2010-2011 Giovanni Mascellani <mascellani@poisson.phc.unipi.it>
# Copyright © 2010-2011 Stefano Maggiolo <s.maggiolo@gmail.com>
# Copyright © 2010-2011 Matteo Boscariol <boscarim@hotmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Submission-related database interface for SQLAlchemy. Not to be
used directly (import  from SQLAlchemyAll).

"""

import time

from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.collections import column_mapped_collection
from sqlalchemy.ext.orderinglist import ordering_list

from cms.db.SQLAlchemyUtils import Base
from cms.db.Task import Task
from cms.db.User import User


class Submission(Base):
    """Class to store a submission. Not to be used directly (import it
    from SQLAlchemyAll).

    """
    __tablename__ = 'submissions'

    # Auto increment primary key.
    id = Column(Integer, primary_key=True)

    # User (id and object) that did the submission.
    user_id = Column(Integer,
                     ForeignKey(User.id,
                                onupdate="CASCADE", ondelete="CASCADE"),
                     nullable=False)
    user = relationship(User,
                        backref=backref("submissions",
                                        single_parent=True,
                                        cascade="all, delete, delete-orphan"))

    # Task (id and object) of the submission.
    task_id = Column(Integer,
                     ForeignKey(Task.id,
                                onupdate="CASCADE", ondelete="CASCADE"),
                     nullable=False)
    task = relationship(Task)

    # Time of the submission.
    timestamp = Column(Integer, nullable=False)

    # Language of submission, or None if not applicable.
    language = Column(String, nullable=True)

    # Compilation outcome (can be None = yet to compile, "ok" =
    # compilation successful and we can evaluate, "fail" =
    # compilation unsuccessful, thorow it away).
    compilation_outcome = Column(String, nullable=True)

    # String containing output from the sandbox, and the compiler
    # stdout and stderr.
    compilation_text = Column(String, nullable=True)

    # Number of tentatives of compilation.
    compilation_tries = Column(Integer, nullable=False)

    # Number of tentatives of evaluation.
    evaluation_tries = Column(Integer, nullable=False)

    # Follows the description of the fields automatically added by
    # SQLAlchemy.
    # files (dict of File objects indexed by filename)
    # executables (dict of Executable objects indexed by filename)
    # evaluations (list of Evaluation objects, one for testcase)
    # token (Token object or None)

    LANGUAGES = ["c", "cpp", "pas"]

    def __init__(self, user, task, timestamp, files, language=None,
                 compilation_outcome=None, compilation_text=None,
                 compilation_tries=0, executables=None,
                 evaluation_tries=0, evaluations=None,
                 token=None):
        self.user = user
        self.task = task
        self.timestamp = timestamp
        self.files = files
        self.language = language
        self.compilation_outcome = compilation_outcome
        if executables is None:
            executables = {}
        self.executables = executables
        self.compilation_text = compilation_text
        if evaluations is None:
            evaluations = []
        self.evaluations = evaluations
        self.compilation_tries = compilation_tries
        self.evaluation_tries = evaluation_tries
        self.token = token

    def export_to_dict(self):
        """Return object data as a dictionary.

        """
        res = {'task':                self.task.num,
               'timestamp':           self.timestamp,
               'files':               [file.export_to_dict() for file in self.files.itervalues()],
               'language':            self.language,
               'compilation_outcome': self.compilation_outcome,
               'compilation_tries':   self.compilation_tries,
               'compilation_text':    self.compilation_text,
               'executables':         [executable.export_to_dict() for executable in self.executables.itervalues()],
               'evaluations':         [evaluation.export_to_dict() for evaluation in self.evaluations],
               'evaluation_tries':    self.evaluation_tries,
               'token':               self.token}
        if self.token is not None:
            res['token'] = self.token.export_to_dict()
        return res

    def tokened(self):
        """Return if the user played a token against the submission.

        return (bool): True if tokened, False otherwise.

        """
        return self.token is not None

    def evaluated(self):
        """Return if the submission has been evaluated.

        return (bool): True if evaluated, False otherwise.

        """
        return self.evaluations != []

    def invalid(self):
        """Blank all compilation and evaluation outcomes, so that ES
        will reschedule the submission for compilation.

        """
        self.compilation_outcome = None
        self.compilation_text = None
        self.evaluations = []
        self.executables = {}

    def invalid_evaluation(self):
        """Blank only the evaluation outcomes, so ES will reschedule
        the submission for evaluation.

        """
        self.evaluations = []

    def play_token(self, timestamp=None):
        """Tell the submission that a token has been used.

        timestamp (int): the time the token has been played.

        """
        self.token = Token(timestamp=timestamp)


class Token(Base):
    """Class to store information about a token. Not to be used
    directly (import it from SQLAlchemyAll).

    """
    __tablename__ = 'tokens'

    # Auto increment primary key.
    id = Column(Integer, primary_key=True)

    # Submission (id and object) the token has been played against.
    submission_id = Column(Integer,
                           ForeignKey(Submission.id,
                                      onupdate="CASCADE", ondelete="CASCADE"),
                           nullable=False)
    submission = relationship(Submission,
                              backref=backref("token",
                                              uselist=False,
                                              single_parent=True,
                                              cascade="all, delete, delete-orphan"),
                              single_parent=True)

    # Time the token was played.
    timestamp = Column(Integer, nullable=False)

    def __init__(self, timestamp=None, submission=None):
        if timestamp is None:
            timestamp = int(time.time())
        self.timestamp = timestamp
        self.submission = submission

    def export_to_dict(self):
        """Return object data as a dictionary.

        """
        return {'timestamp': self.timestamp}


class File(Base):
    """Class to store information about one file submitted within a
    submission. Not to be used directly (import it from
    SQLAlchemyAll).

    """
    __tablename__ = 'files'

    # Auto increment primary key.
    id = Column(Integer, primary_key=True)

    # Filename and digest of the submitted file.
    filename = Column(String, nullable=False)
    digest = Column(String, nullable=False)

    # Submission (id and object) of the submission.
    submission_id = Column(Integer,
                           ForeignKey(Submission.id,
                                      onupdate="CASCADE", ondelete="CASCADE"),
                           nullable=False)
    submission = relationship(
        Submission,
        backref=backref('files',
                        collection_class=column_mapped_collection(filename),
                        single_parent=True,
                        cascade="all, delete, delete-orphan"))

    def __init__(self, digest, filename=None, submission=None):
        self.filename = filename
        self.digest = digest
        self.submission = submission

    def export_to_dict(self):
        """Return object data as a dictionary.

        """
        return {'filename': self.filename,
                'digest':   self.digest}


class Executable(Base):
    """Class to store information about one file generated by the
    compilation of a submission. Not to be used directly (import it
    from SQLAlchemyAll).

    """
    __tablename__ = 'executables'

    # Auto increment primary key.
    id = Column(Integer, primary_key=True)

    # Filename and digest of the file.
    filename = Column(String, nullable=False)
    digest = Column(String, nullable=False)

    # Submission (id and object) of the submission.
    submission_id = Column(Integer,
                           ForeignKey(Submission.id,
                                      onupdate="CASCADE", ondelete="CASCADE"),
                           nullable=False)
    submission = relationship(
        Submission,
        backref=backref('executables',
                        collection_class=column_mapped_collection(filename),
                        single_parent=True,
                        cascade="all, delete, delete-orphan"))

    def __init__(self, digest, filename=None, submission=None):
        self.filename = filename
        self.digest = digest
        self.submission = submission

    def export_to_dict(self):
        """Return object data as a dictionary.

        """
        return {'filename': self.filename,
                'digest':   self.digest}


class Evaluation(Base):
    """Class to store information about the outcome of the evaluation
    of a submission against one testcase. Not to be used directly
    (import it from SQLAlchemyAll).

    """
    __tablename__ = 'evaluations'

    # Auto increment primary key.
    id = Column(Integer, primary_key=True)

    # Number of the testcase
    num = Column(Integer, nullable=False)

    # Submission (id and object) of the submission.
    submission_id = Column(Integer,
                           ForeignKey(Submission.id,
                                      onupdate="CASCADE", ondelete="CASCADE"),
                           nullable=False)
    submission = relationship(
        Submission,
        backref=backref('evaluations',
                        collection_class=ordering_list('num'),
                        order_by=[num],
                        single_parent=True,
                        cascade="all, delete, delete-orphan"))

    # String containing output from the grader (usually "Correct",
    # "Time limit", ...).
    text = Column(String, nullable=True)

    # String containing the outcome of the evaluation (usually 1.0,
    # ...) not necessary the points awarded, that will be computed by
    # the scorer.
    outcome = Column(String, nullable=True)

    def __init__(self, text, outcome, num=None, submission=None):
        self.text = text
        self.outcome = outcome
        self.num = num
        self.submission = submission

    def export_to_dict(self):
        """Return object data as a dictionary.

        """
        return {'text':    self.text,
                'outcome': self.outcome,
                'num':     self.num}
